# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import time
import threading
from typing import Generator

import pytest

from instana.autoprofile.profiler import Profiler
from instana.autoprofile.runtime import RuntimeInfo
from instana.autoprofile.samplers.cpu_sampler import CPUSampler


class TestCPUSampler:
    @pytest.fixture(autouse=True)
    def _resources(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Create a new Profiler.
        self.profiler = Profiler(None)
        self.profiler.start(disable_timers=True)
        yield
        # teardown
        self.profiler.destroy()


    def test_cpu_profile(self) -> None:
        if RuntimeInfo.OS_WIN:
            return

        sampler = CPUSampler(self.profiler)
        sampler.setup()
        sampler.reset()

        def record() -> None:
            sampler.start_sampler()
            time.sleep(2)
            sampler.stop_sampler()

        record_t = threading.Thread(target=record)
        record_t.start()

        def cpu_work_main_thread() -> None:
            for i in range(0, 1000000):
                text = "text1" + str(i)
                text = text + "text2"

        cpu_work_main_thread()

        record_t.join()

        profile = sampler.build_profile(2000, 120000).to_dict()

        assert 'cpu_work_main_thread' in str(profile)



# (c) Copyright IBM Corp. 2024

from fastapi import FastAPI, HTTPException, Response
from fastapi.concurrency import run_in_threadpool
from fastapi.middleware import Middleware
from fastapi.middleware.trustedhost import TrustedHostMiddleware


fastapi_server = FastAPI(
    middleware=[
        Middleware(
            TrustedHostMiddleware,
            allowed_hosts=["*"],
        ),
    ],
)


@fastapi_server.get("/")
async def root():
    return {"message": "Hello World"}



# (c) Copyright IBM Corp. 2024

import logging
import time
from typing import Generator
from unittest.mock import patch

import pytest
from opentelemetry.trace.status import Status, StatusCode

from instana.recorder import StanRecorder
from instana.span.span import INVALID_SPAN, Event, InstanaSpan, get_current_span
from instana.span_context import SpanContext


class TestSpan:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        self.span = None
        yield
        if isinstance(self.span, InstanaSpan):
            self.span.events.clear()

    def test_span_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        trace_id: int,
        span_id: int,
    ) -> None:
        span_name = "test-span"
        timestamp = time.time_ns()
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert self.span is not None
        assert isinstance(self.span, InstanaSpan)
        assert self.span.name == span_name

        context = self.span.context
        assert isinstance(context, SpanContext)
        assert context.trace_id == trace_id
        assert context.span_id == span_id

        assert self.span.start_time
        assert isinstance(self.span.start_time, int)
        assert self.span.start_time > timestamp
        assert not self.span.end_time
        assert not self.span.attributes
        assert not self.span.events
        assert self.span.is_recording()
        assert self.span.status
        assert self.span.status.is_unset

    def test_span_get_span_context(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        trace_id: int,
        span_id: int,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        context = self.span.get_span_context()
        assert isinstance(context, SpanContext)
        assert context.trace_id == trace_id
        assert context.span_id == span_id
        assert context == self.span.context

    def test_span_set_attributes_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert not self.span.attributes

        attributes = {
            "field1": 1,
            "field2": "two",
        }
        self.span.set_attributes(attributes)

        assert self.span.attributes
        assert len(self.span.attributes) == 2
        assert "field1" in self.span.attributes.keys()
        assert "two" == self.span.attributes.get("field2")

    def test_span_set_attributes(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        attributes = {
            "field1": 1,
            "field2": "two",
        }
        self.span = InstanaSpan(
            span_name, span_context, span_processor, attributes=attributes
        )

        assert self.span.attributes
        assert len(self.span.attributes) == 2
        assert "field1" in self.span.attributes.keys()
        assert "two" == self.span.attributes.get("field2")

        attributes = {
            "field3": True,
            "field4": ["four", "vier", "quatro"],
        }
        self.span.set_attributes(attributes)

        assert len(self.span.attributes) == 4
        assert "field3" in self.span.attributes.keys()
        assert "vier" in self.span.attributes.get("field4")

    def test_span_set_attribute_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert not self.span.attributes

        attributes = {
            "field1": 1,
            "field2": "two",
        }
        for key, value in attributes.items():
            self.span.set_attribute(key, value)

        assert self.span.attributes
        assert len(self.span.attributes) == 2
        assert "field1" in self.span.attributes.keys()
        assert "two" == self.span.attributes.get("field2")

    def test_span_set_attribute(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        attributes = {
            "field1": 1,
            "field2": "two",
        }
        self.span = InstanaSpan(
            span_name, span_context, span_processor, attributes=attributes
        )

        assert self.span.attributes
        assert len(self.span.attributes) == 2
        assert "field1" in self.span.attributes.keys()
        assert "two" == self.span.attributes.get("field2")

        attributes = {
            "field3": True,
            "field4": ["four", "vier", "quatro"],
        }
        for key, value in attributes.items():
            self.span.set_attribute(key, value)

        assert len(self.span.attributes) == 4
        assert "field3" in self.span.attributes.keys()
        assert "vier" in self.span.attributes.get("field4")

    def test_span_update_name(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span-1"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert self.span is not None
        assert isinstance(self.span, InstanaSpan)
        assert self.span.name == span_name

        new_span_name = "test-span-2"
        self.span.update_name(new_span_name)
        assert self.span is not None
        assert isinstance(self.span, InstanaSpan)
        assert self.span.name == new_span_name

    def test_span_set_status_with_Status_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.WARNING, logger="opentelemetry.trace.status")
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert self.span.status
        assert self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code == StatusCode.UNSET
        assert self.span.status.status_code != StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

        status_desc = "Status is OK."
        span_status = Status(status_code=StatusCode.OK, description=status_desc)

        assert (
            "description should only be set when status_code is set to StatusCode.ERROR"
            in caplog.messages
        )

        self.span.set_status(span_status)

        assert self.span.status
        assert not self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code == StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

    def test_span_set_status_with_Status_and_desc(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.WARNING, logger="opentelemetry.trace.status")
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert self.span.status
        assert self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code == StatusCode.UNSET
        assert self.span.status.status_code != StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

        status_desc = "Status is OK."
        span_status = Status(status_code=StatusCode.OK, description=status_desc)

        assert (
            "description should only be set when status_code is set to StatusCode.ERROR"
            in caplog.messages
        )

        set_status_desc = "Test"
        self.span.set_status(span_status, set_status_desc)
        excepted_log = f"Description {set_status_desc} ignored. Use either `Status` or `(StatusCode, Description)`"

        assert self.span.status
        assert not self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert excepted_log in caplog.messages
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code == StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

    def test_span_set_status_with_StatusUNSET_to_StatusERROR(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.WARNING, logger="opentelemetry.trace.status")
        span_name = "test-span"
        status_desc = "Status is UNSET."
        span_status = Status(status_code=StatusCode.UNSET, description=status_desc)

        assert (
            "description should only be set when status_code is set to StatusCode.ERROR"
            in caplog.messages
        )

        self.span = InstanaSpan(
            span_name, span_context, span_processor, status=span_status
        )

        assert self.span.status
        assert self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code == StatusCode.UNSET
        assert self.span.status.status_code != StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

        status_desc = "Houston we have a problem!"
        span_status = Status(StatusCode.ERROR, status_desc)
        self.span.set_status(span_status)

        assert self.span.status
        assert not self.span.status.is_unset
        assert not self.span.status.is_ok
        assert self.span.status.description == status_desc
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code != StatusCode.OK
        assert self.span.status.status_code == StatusCode.ERROR

    def test_span_set_status_with_StatusOK_to_StatusERROR(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.WARNING, logger="opentelemetry.trace.status")

        span_name = "test-span"
        status_desc = "Status is OK."
        span_status = Status(status_code=StatusCode.OK, description=status_desc)

        assert (
            "description should only be set when status_code is set to StatusCode.ERROR"
            in caplog.messages
        )

        self.span = InstanaSpan(
            span_name, span_context, span_processor, status=span_status
        )

        assert self.span.status
        assert not self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code == StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

        status_desc = "Houston we have a problem!"
        span_status = Status(StatusCode.ERROR, status_desc)
        self.span.set_status(span_status)

        assert self.span.status
        assert not self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code == StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

    def test_span_set_status_with_StatusCode_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert self.span.status
        assert self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code == StatusCode.UNSET
        assert self.span.status.status_code != StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

        span_status_code = StatusCode(StatusCode.OK)

        self.span.set_status(span_status_code)

        assert self.span.status
        assert not self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code == StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

    def test_span_set_status_with_StatusCode_and_desc(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.WARNING, logger="opentelemetry.trace.status")

        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert self.span.status
        assert self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code == StatusCode.UNSET
        assert self.span.status.status_code != StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

        status_desc = "Status is OK."
        span_status_code = StatusCode(StatusCode.OK)
        self.span.set_status(span_status_code, status_desc)

        assert self.span.status
        assert not self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code == StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR
        assert (
            "description should only be set when status_code is set to StatusCode.ERROR"
            in caplog.messages
        )

    def test_span_set_status_with_StatusCodeUNSET_to_StatusCodeERROR(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.WARNING, logger="opentelemetry.trace.status")

        span_name = "test-span"
        status_desc = "Status is UNSET."
        span_status = Status(status_code=StatusCode.UNSET, description=status_desc)

        assert (
            "description should only be set when status_code is set to StatusCode.ERROR"
            in caplog.messages
        )

        self.span = InstanaSpan(
            span_name, span_context, span_processor, status=span_status
        )

        assert self.span.status
        assert self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code == StatusCode.UNSET
        assert self.span.status.status_code != StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

        status_desc = "Houston we have a problem!"
        span_status_code = StatusCode(StatusCode.ERROR)
        self.span.set_status(span_status_code, status_desc)

        assert self.span.status
        assert not self.span.status.is_unset
        assert not self.span.status.is_ok
        assert self.span.status.description == status_desc
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code != StatusCode.OK
        assert self.span.status.status_code == StatusCode.ERROR

    def test_span_set_status_with_StatusCodeOK_to_StatusCodeERROR(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        caplog.set_level(logging.WARNING, logger="opentelemetry.trace.status")

        span_name = "test-span"
        status_desc = "Status is OK."
        span_status = Status(status_code=StatusCode.OK, description=status_desc)

        assert (
            "description should only be set when status_code is set to StatusCode.ERROR"
            in caplog.messages
        )

        self.span = InstanaSpan(
            span_name, span_context, span_processor, status=span_status
        )

        assert self.span.status
        assert not self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code == StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

        status_desc = "Houston we have a problem!"
        span_status_code = StatusCode(StatusCode.ERROR)
        self.span.set_status(span_status_code, status_desc)

        assert self.span.status
        assert not self.span.status.is_unset
        assert self.span.status.is_ok
        assert not self.span.status.description
        assert self.span.status.status_code != StatusCode.UNSET
        assert self.span.status.status_code == StatusCode.OK
        assert self.span.status.status_code != StatusCode.ERROR

    def test_span_add_event_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert not self.span.events

        event_name = "event1"
        attributes = {
            "field1": 1,
            "field2": "two",
        }
        timestamp = time.time_ns()
        self.span.add_event(event_name, attributes, timestamp)

        assert self.span.events
        assert len(self.span.events) == 1
        for event in self.span.events:
            assert isinstance(event, Event)
            assert event.name == event_name
            assert event.timestamp == timestamp
            assert len(event.attributes) == 2

    def test_span_add_event(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        event_name1 = "event1"
        attributes = {
            "field1": 1,
            "field2": "two",
        }
        timestamp1 = time.time_ns()
        event = Event(event_name1, attributes, timestamp1)
        self.span = InstanaSpan(span_name, span_context, span_processor, events=[event])

        assert self.span.events
        assert len(self.span.events) == 1
        for event in self.span.events:
            assert isinstance(event, Event)
            assert event.name == event_name1
            assert event.timestamp == timestamp1
            assert len(event.attributes) == 2

        event_name2 = "event2"
        attributes = {
            "field3": True,
            "field4": ["four", "vier", "quatro"],
        }
        timestamp2 = time.time_ns()
        self.span.add_event(event_name2, attributes, timestamp2)

        assert len(self.span.events) == 2
        for event in self.span.events:
            assert isinstance(event, Event)
            assert event.name in [event_name1, event_name2]
            assert event.timestamp in [timestamp1, timestamp2]
            assert len(event.attributes) == 2

    @pytest.mark.parametrize(
        "span_name, span_attribute",
        [
            ("test-span", None),
            ("rpc-server", "rpc.error"),
            ("rpc-client", "rpc.error"),
            ("mysql", "mysql.error"),
            ("postgres", "pg.error"),
            ("django", "http.error"),
            ("http", "http.error"),
            ("urllib3", "http.error"),
            ("wsgi", "http.error"),
            ("asgi", "http.error"),
            ("celery-client", "error"),
            ("celery-worker", "error"),
            ("sqlalchemy", "sqlalchemy.err"),
            ("aws.lambda.entry", "lambda.error"),
        ],
    )
    def test_span_record_exception_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
        span_name: str,
        span_attribute: str,
    ) -> None:
        exception_msg = "Test Exception"

        exception = Exception(exception_msg)
        self.span = InstanaSpan(span_name, span_context, span_processor)

        self.span.record_exception(exception)

        assert span_name == self.span.name
        assert 1 == self.span.attributes.get("ec", 0)
        if span_attribute:
            assert span_attribute in self.span.attributes.keys()
            assert exception_msg == self.span.attributes.get(span_attribute, None)
        else:
            event = self.span.events[-1]  # always get the latest event
            assert isinstance(event, Event)
            assert "exception" == event.name
            assert exception_msg == event.attributes.get("message", None)

    def test_span_record_exception_with_attribute(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        exception_msg = "Test Exception"
        attributes = {
            "custom_attr": 0,
        }

        exception = Exception(exception_msg)
        self.span = InstanaSpan(span_name, span_context, span_processor)

        self.span.record_exception(exception, attributes)

        assert span_name == self.span.name
        assert 1 == self.span.attributes.get("ec", 0)

        event = self.span.events[-1]  # always get the latest event
        assert isinstance(event, Event)
        assert 2 == len(event.attributes)
        assert exception_msg == event.attributes.get("message", None)
        assert 0 == event.attributes.get("custom_attr", None)

    def test_span_record_exception_with_Exception_msg(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "wsgi"
        span_attribute = "http.error"
        exception_msg = "Test Exception"

        exception = Exception()
        exception.message = exception_msg
        self.span = InstanaSpan(span_name, span_context, span_processor)

        self.span.record_exception(exception)

        assert span_name == self.span.name
        assert 1 == self.span.attributes.get("ec", 0)
        assert span_attribute in self.span.attributes.keys()
        assert exception_msg == self.span.attributes.get(span_attribute, None)

    def test_span_record_exception_with_Exception_none_msg(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "wsgi"
        span_attribute = "http.error"

        exception = Exception()
        exception.message = None
        self.span = InstanaSpan(span_name, span_context, span_processor)

        self.span.record_exception(exception)

        assert span_name == self.span.name
        assert 1 == self.span.attributes.get("ec", 0)
        assert span_attribute in self.span.attributes.keys()
        assert "Exception()" == self.span.attributes.get(span_attribute, None)

    def test_span_record_exception_with_Exception_raised(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"

        exception = None
        self.span = InstanaSpan(span_name, span_context, span_processor)

        with patch(
            "instana.span.span.InstanaSpan.add_event",
            side_effect=Exception("mocked error"),
        ):
            with pytest.raises(Exception):
                self.span.record_exception(exception)

    def test_span_end_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert not self.span.end_time

        self.span.end()

        assert self.span.end_time
        assert isinstance(self.span.end_time, int)

    def test_span_end(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"

        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert not self.span.end_time

        timestamp_end = time.time_ns()
        self.span.end(timestamp_end)

        assert self.span.end_time
        assert self.span.end_time == timestamp_end

    def test_span_mark_as_errored_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        attributes = {
            "ec": 0,
        }
        self.span = InstanaSpan(
            span_name, span_context, span_processor, attributes=attributes
        )

        assert self.span.attributes
        assert len(self.span.attributes) == 1
        assert self.span.attributes.get("ec") == 0

        self.span.mark_as_errored()

        assert self.span.attributes
        assert len(self.span.attributes) == 1
        assert self.span.attributes.get("ec") == 1

    def test_span_mark_as_errored(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        attributes = {
            "ec": 0,
        }
        self.span = InstanaSpan(
            span_name, span_context, span_processor, attributes=attributes
        )

        assert self.span.attributes
        assert len(self.span.attributes) == 1
        assert self.span.attributes.get("ec") == 0

        attributes = {
            "field1": 1,
            "field2": "two",
        }
        self.span.mark_as_errored(attributes)

        assert self.span.attributes
        assert len(self.span.attributes) == 3
        assert self.span.attributes.get("ec") == 1
        assert "field1" in self.span.attributes.keys()
        assert self.span.attributes.get("field2") == "two"

        self.span.mark_as_errored()

        assert self.span.attributes
        assert len(self.span.attributes) == 3
        assert self.span.attributes.get("ec") == 2
        assert "field1" in self.span.attributes.keys()
        assert self.span.attributes.get("field2") == "two"

    def test_span_mark_as_errored_exception(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        with patch(
            "instana.span.span.InstanaSpan.set_attribute",
            side_effect=Exception("mocked error"),
        ):
            self.span.mark_as_errored()
            assert not self.span.attributes

    def test_span_assure_errored_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        self.span.assure_errored()

        assert self.span.attributes
        assert len(self.span.attributes) == 1
        assert self.span.attributes.get("ec") == 1

    def test_span_assure_errored(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        attributes = {
            "ec": 0,
        }
        self.span = InstanaSpan(
            span_name, span_context, span_processor, attributes=attributes
        )

        assert self.span.attributes
        assert len(self.span.attributes) == 1
        assert self.span.attributes.get("ec") == 0

        self.span.assure_errored()

        assert self.span.attributes
        assert len(self.span.attributes) == 1
        assert self.span.attributes.get("ec") == 1

    def test_span_assure_errored_exception(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        with patch(
            "instana.span.span.InstanaSpan.set_attribute",
            side_effect=Exception("mocked error"),
        ):
            self.span.assure_errored()
            assert not self.span.attributes

    def test_get_current_span(self, context: SpanContext) -> None:
        self.span = get_current_span(context)
        assert isinstance(self.span, InstanaSpan)

    def test_get_current_span_INVALID_SPAN(self) -> None:
        self.span = get_current_span()

        assert self.span
        assert self.span == INVALID_SPAN

    def test_span_duration_default(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert not self.span.end_time
        assert not self.span.duration

        self.span.end()

        assert self.span.end_time
        assert self.span.duration
        assert isinstance(self.span.duration, int)
        assert self.span.duration > 0

    def test_span_duration(
        self,
        span_context: SpanContext,
        span_processor: StanRecorder,
    ) -> None:
        span_name = "test-span"
        self.span = InstanaSpan(span_name, span_context, span_processor)

        assert not self.span.end_time
        assert not self.span.duration

        timestamp_end = time.time_ns()
        self.span.end(timestamp_end)

        assert self.span.end_time
        assert self.span.end_time == timestamp_end
        assert self.span.duration
        assert isinstance(self.span.duration, int)
        assert self.span.duration > 0
        assert self.span.duration == (timestamp_end - self.span.start_time)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import sys
import MySQLdb
import pytest

from instana.singletons import agent, tracer
from tests.helpers import testenv


@pytest.mark.skipif(
    sys.platform == "darwin",
    reason="Avoiding errors with deprecated MySQL Client lib.",
)
class TestMySQLPython:
    @pytest.fixture(autouse=True)
    def _resource(self):
        self.db = MySQLdb.connect(
            host=testenv["mysql_host"],
            port=testenv["mysql_port"],
            user=testenv["mysql_user"],
            passwd=testenv["mysql_pw"],
            db=testenv["mysql_db"],
        )
        database_setup_query = """
        DROP TABLE IF EXISTS users;
        CREATE TABLE users(
            id serial primary key,
            name varchar(40) NOT NULL,
            email varchar(40) NOT NULL
        );
        INSERT INTO users(name, email) VALUES('kermit', 'kermit@muppets.com');
        DROP PROCEDURE IF EXISTS test_proc;
        CREATE PROCEDURE test_proc(IN t VARCHAR(255))
        BEGIN
            SELECT name FROM users WHERE name = t;
        END
        """
        setup_cursor = self.db.cursor()
        setup_cursor.execute(database_setup_query)
        setup_cursor.close()

        self.cursor = self.db.cursor()
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        tracer.cur_ctx = None
        yield
        if self.cursor and self.cursor.connection.open:
            self.cursor.close()
        if self.db and self.db.open:
            self.db.close()
        agent.options.allow_exit_as_root = False

    def test_vanilla_query(self):
        affected_rows = self.cursor.execute("""SELECT * from users""")
        assert affected_rows == 1
        result = self.cursor.fetchone()
        assert len(result) == 3

        spans = self.recorder.queued_spans()
        assert len(spans) == 0

    def test_basic_query(self):
        with tracer.start_as_current_span("test"):
            affected_rows = self.cursor.execute("""SELECT * from users""")
            result = self.cursor.fetchone()

        assert affected_rows == 1
        assert len(result) == 3

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mysql"
        assert db_span.data["mysql"]["db"] == testenv["mysql_db"]
        assert db_span.data["mysql"]["user"] == testenv["mysql_user"]
        assert db_span.data["mysql"]["stmt"] == "SELECT * from users"
        assert db_span.data["mysql"]["host"] == testenv["mysql_host"]
        assert db_span.data["mysql"]["port"] == testenv["mysql_port"]

    def test_basic_query_as_root_exit_span(self):
        agent.options.allow_exit_as_root = True
        affected_rows = self.cursor.execute("""SELECT * from users""")
        result = self.cursor.fetchone()

        assert affected_rows == 1
        assert len(result) == 3

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        db_span = spans[0]

        assert not db_span.ec

        assert db_span.n == "mysql"
        assert db_span.data["mysql"]["db"] == testenv["mysql_db"]
        assert db_span.data["mysql"]["user"] == testenv["mysql_user"]
        assert db_span.data["mysql"]["stmt"] == "SELECT * from users"
        assert db_span.data["mysql"]["host"] == testenv["mysql_host"]
        assert db_span.data["mysql"]["port"] == testenv["mysql_port"]

    def test_basic_insert(self):
        with tracer.start_as_current_span("test"):
            affected_rows = self.cursor.execute(
                """INSERT INTO users(name, email) VALUES(%s, %s)""",
                ("beaker", "beaker@muppets.com"),
            )

        assert affected_rows == 1

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mysql"
        assert db_span.data["mysql"]["db"] == testenv["mysql_db"]
        assert db_span.data["mysql"]["user"] == testenv["mysql_user"]
        assert (
            db_span.data["mysql"]["stmt"]
            == "INSERT INTO users(name, email) VALUES(%s, %s)"
        )
        assert db_span.data["mysql"]["host"] == testenv["mysql_host"]
        assert db_span.data["mysql"]["port"] == testenv["mysql_port"]

    def test_executemany(self):
        with tracer.start_as_current_span("test"):
            affected_rows = self.cursor.executemany(
                "INSERT INTO users(name, email) VALUES(%s, %s)",
                [("beaker", "beaker@muppets.com"), ("beaker", "beaker@muppets.com")],
            )
            self.db.commit()

        assert affected_rows == 2

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mysql"
        assert db_span.data["mysql"]["db"] == testenv["mysql_db"]
        assert db_span.data["mysql"]["user"] == testenv["mysql_user"]
        assert (
            db_span.data["mysql"]["stmt"]
            == "INSERT INTO users(name, email) VALUES(%s, %s)"
        )
        assert db_span.data["mysql"]["host"] == testenv["mysql_host"]
        assert db_span.data["mysql"]["port"] == testenv["mysql_port"]

    def test_call_proc(self):
        with tracer.start_as_current_span("test"):
            callproc_result = self.cursor.callproc("test_proc", ("beaker",))

        assert isinstance(callproc_result, tuple)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mysql"
        assert db_span.data["mysql"]["db"] == testenv["mysql_db"]
        assert db_span.data["mysql"]["user"] == testenv["mysql_user"]
        assert db_span.data["mysql"]["stmt"] == "test_proc"
        assert db_span.data["mysql"]["host"] == testenv["mysql_host"]
        assert db_span.data["mysql"]["port"] == testenv["mysql_port"]

    def test_error_capture(self):
        affected_rows = None
        try:
            with tracer.start_as_current_span("test"):
                affected_rows = self.cursor.execute("""SELECT * from blah""")
        except Exception:
            pass

        assert not affected_rows

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert db_span.ec == 2
        assert (
            db_span.data["mysql"]["error"]
            == f"(1146, \"Table '{testenv['mysql_db']}.blah' doesn't exist\")"
        )

        assert db_span.n == "mysql"
        assert db_span.data["mysql"]["db"] == testenv["mysql_db"]
        assert db_span.data["mysql"]["user"] == testenv["mysql_user"]
        assert db_span.data["mysql"]["stmt"] == "SELECT * from blah"
        assert db_span.data["mysql"]["host"] == testenv["mysql_host"]
        assert db_span.data["mysql"]["port"] == testenv["mysql_port"]

    def test_connect_cursor_ctx_mgr(self):
        with tracer.start_as_current_span("test"):
            with self.db as connection:
                with connection.cursor() as cursor:
                    affected_rows = cursor.execute("""SELECT * from users""")

        assert affected_rows == 1
        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mysql"
        assert db_span.data["mysql"]["db"] == testenv["mysql_db"]
        assert db_span.data["mysql"]["user"] == testenv["mysql_user"]
        assert db_span.data["mysql"]["stmt"] == "SELECT * from users"
        assert db_span.data["mysql"]["host"] == testenv["mysql_host"]
        assert db_span.data["mysql"]["port"] == testenv["mysql_port"]

    def test_connect_ctx_mgr(self):
        with tracer.start_as_current_span("test"):
            with self.db as connection:
                cursor = connection.cursor()
                cursor.execute("""SELECT * from users""")

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mysql"
        assert db_span.data["mysql"]["db"] == testenv["mysql_db"]
        assert db_span.data["mysql"]["user"] == testenv["mysql_user"]
        assert db_span.data["mysql"]["stmt"] == "SELECT * from users"
        assert db_span.data["mysql"]["host"] == testenv["mysql_host"]
        assert db_span.data["mysql"]["port"] == testenv["mysql_port"]

    def test_cursor_ctx_mgr(self):
        with tracer.start_as_current_span("test"):
            connection = self.db
            with connection.cursor() as cursor:
                affected_rows = cursor.execute("""SELECT * from users""")

        assert affected_rows == 1
        spans = self.recorder.queued_spans()
        assert len(spans) == 2

        db_span, test_span = spans

        assert test_span.data["sdk"]["name"] == "test"
        assert test_span.t == db_span.t
        assert db_span.p == test_span.s

        assert not db_span.ec

        assert db_span.n == "mysql"
        assert db_span.data["mysql"]["db"] == testenv["mysql_db"]
        assert db_span.data["mysql"]["user"] == testenv["mysql_user"]
        assert db_span.data["mysql"]["stmt"] == "SELECT * from users"
        assert db_span.data["mysql"]["host"] == testenv["mysql_host"]
        assert db_span.data["mysql"]["port"] == testenv["mysql_port"]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

"""
Instana WSGI Middleware
"""

from typing import Dict, Any, Callable, List, Tuple, Optional

from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry import context, trace

from instana.propagators.format import Format
from instana.singletons import agent, tracer
from instana.util.secrets import strip_secrets_from_query
from instana.util.traceutils import extract_custom_headers


class InstanaWSGIMiddleware(object):
    """Instana WSGI middleware"""

    def __init__(self, app: object) -> None:
        self.app = app

    def __call__(self, environ: Dict[str, Any], start_response: Callable) -> object:
        env = environ

        def new_start_response(
            status: str,
            headers: List[Tuple[object, ...]],
            exc_info: Optional[Exception] = None,
        ) -> object:
            """Modified start response with additional headers."""
            extract_custom_headers(self.span, headers)

            tracer.inject(self.span.context, Format.HTTP_HEADERS, headers)

            headers_str = [
                (header[0], str(header[1]))
                if not isinstance(header[1], str)
                else header
                for header in headers
            ]
            res = start_response(status, headers_str, exc_info)

            sc = status.split(" ")[0]
            if 500 <= int(sc):
                self.span.mark_as_errored()

            self.span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, sc)
            if self.span and self.span.is_recording():
                self.span.end()
            if self.token:
                context.detach(self.token)
            return res

        span_context = tracer.extract(Format.HTTP_HEADERS, env)
        self.span = tracer.start_span("wsgi", span_context=span_context)

        ctx = trace.set_span_in_context(self.span)
        self.token = context.attach(ctx)

        extract_custom_headers(self.span, env, format=True)

        if "PATH_INFO" in env:
            self.span.set_attribute("http.path", env["PATH_INFO"])
        if "QUERY_STRING" in env and len(env["QUERY_STRING"]):
            scrubbed_params = strip_secrets_from_query(
                env["QUERY_STRING"],
                agent.options.secrets_matcher,
                agent.options.secrets_list,
            )
            self.span.set_attribute("http.params", scrubbed_params)
        if "REQUEST_METHOD" in env:
            self.span.set_attribute(SpanAttributes.HTTP_METHOD, env["REQUEST_METHOD"])
        if "HTTP_HOST" in env:
            self.span.set_attribute("http.host", env["HTTP_HOST"])

        return self.app(environ, new_start_response)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import random
import threading
import time
from typing import Generator, Optional

import pytest

from instana.autoprofile.profiler import Profiler
from instana.autoprofile.runtime import min_version, RuntimeInfo
from instana.autoprofile.samplers.allocation_sampler import AllocationSampler


class TestAllocationSampler:
    @pytest.fixture(autouse=True)
    def _resources(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Create a new Profiler.
        self.profiler = Profiler(None)
        self.profiler.start(disable_timers=True)
        yield
        # teardown
        self.profiler.destroy()

    def test_allocation_profile(self) -> None:
        if RuntimeInfo.OS_WIN or not min_version(3, 4):
            return

        sampler = AllocationSampler(self.profiler)
        sampler.setup()
        sampler.reset()

        mem1 = []

        def mem_leak(n: Optional[int] = 100000) -> None:
            mem2 = []
            for i in range(0, n):
                mem1.append(random.randint(0, 1000))
                mem2.append(random.randint(0, 1000))

        def mem_leak2() -> None:
            mem_leak()

        def mem_leak3() -> None:
            mem_leak2()

        def mem_leak4() -> None:
            mem_leak3()

        def mem_leak5() -> None:
            mem_leak4()

        def record() -> None:
            sampler.start_sampler()
            time.sleep(2)
            sampler.stop_sampler()

        t = threading.Thread(target=record)
        t.start()

        # simulate leak
        mem_leak5()

        t.join()

        profile = sampler.build_profile(2000, 120000).to_dict()

        assert "test_allocation_sampler.py" in str(profile)




# (c) Copyright IBM Corp. 2024

import os

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.responses import PlainTextResponse
from starlette.routing import Route

dir_path = os.path.dirname(os.path.realpath(__file__))


def homepage(request):
    return PlainTextResponse("Hello, world!")


def five_hundred(request):
    return PlainTextResponse("Something went wrong!", status_code=500)


def startup():
    print("Ready to go")


routes = [
    Route("/", homepage),
    Route("/five", five_hundred),
]

starlette_server = Starlette(
    debug=True,
    routes=routes,
    on_startup=[startup],
    middleware=[
        Middleware(
            TrustedHostMiddleware,
            allowed_hosts=["*"],
        ),
    ],
)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import time
import urllib3
import pytest
from typing import Generator

from instana.util.ids import hex_id
from tests.apps import bottle_app
from tests.helpers import testenv
from instana.singletons import agent, tracer
from instana.span.span import get_current_span


class TestWSGI:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """ Clear all spans before a test run """
        self.http = urllib3.PoolManager()
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        time.sleep(0.1)

    def test_vanilla_requests(self) -> None:
        response = self.http.request('GET', testenv["wsgi_server"] + '/')
        spans = self.recorder.queued_spans()

        assert 1 == len(spans)
        assert get_current_span().is_recording() is False
        assert response.status == 200

    def test_get_request(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["wsgi_server"] + '/')

        spans = self.recorder.queued_spans()

        assert 3 == len(spans)
        assert get_current_span().is_recording() is False

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert 200 == response.status

        assert 'X-INSTANA-T' in response.headers
        assert int(response.headers['X-INSTANA-T'], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert int(response.headers['X-INSTANA-S'], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        assert response.headers['X-INSTANA-L'] == '1'

        assert 'Server-Timing' in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers['Server-Timing'] == server_timing_value

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        assert wsgi_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert wsgi_span.ec is None

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert '127.0.0.1:' + str(testenv["wsgi_port"]) == wsgi_span.data["http"]["host"]
        assert '/' == wsgi_span.data["http"]["path"]
        assert 'GET' == wsgi_span.data["http"]["method"]
        assert "200" == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

    def test_synthetic_request(self) -> None:
        headers = {
            'X-INSTANA-SYNTHETIC': '1'
        }
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["wsgi_server"] + '/', headers=headers)

        spans = self.recorder.queued_spans()

        assert 3 == len(spans)
        assert get_current_span().is_recording() is False

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert wsgi_span.sy
        assert urllib3_span.sy is None
        assert test_span.sy is None

    def test_secret_scrubbing(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["wsgi_server"] + '/?secret=shhh')

        spans = self.recorder.queued_spans()

        assert 3 == len(spans)
        assert get_current_span().is_recording() is False

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert 200 == response.status

        assert 'X-INSTANA-T' in response.headers
        assert int(response.headers['X-INSTANA-T'], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert int(response.headers['X-INSTANA-S'], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        assert response.headers['X-INSTANA-L'] == '1'

        assert 'Server-Timing' in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers['Server-Timing'] == server_timing_value

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert wsgi_span.ec is None

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert '127.0.0.1:' + str(testenv["wsgi_port"]) == wsgi_span.data["http"]["host"]
        assert '/' == wsgi_span.data["http"]["path"]
        assert 'secret=<redacted>' == wsgi_span.data["http"]["params"]
        assert 'GET' == wsgi_span.data["http"]["method"]
        assert "200" == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

    def test_with_incoming_context(self) -> None:
        request_headers = dict()
        request_headers['X-INSTANA-T'] = '0000000000000001'
        request_headers['X-INSTANA-S'] = '0000000000000001'

        response = self.http.request('GET', testenv["wsgi_server"] + '/', headers=request_headers)

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 1 == len(spans)

        wsgi_span = spans[0]

        # assert wsgi_span.t == '0000000000000001'
        # assert wsgi_span.p == '0000000000000001'
        assert wsgi_span.t == 1
        assert wsgi_span.p == 1

        assert 'X-INSTANA-T' in response.headers
        assert int(response.headers['X-INSTANA-T'], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert int(response.headers['X-INSTANA-S'], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        assert response.headers['X-INSTANA-L'] == '1'

        assert 'Server-Timing' in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers['Server-Timing'] == server_timing_value

    def test_with_incoming_mixed_case_context(self) -> None:
        request_headers = dict()
        request_headers['X-InSTANa-T'] = '0000000000000001'
        request_headers['X-instana-S'] = '0000000000000001'

        response = self.http.request('GET', testenv["wsgi_server"] + '/', headers=request_headers)

        assert response
        assert 200 == response.status

        spans = self.recorder.queued_spans()
        assert 1 == len(spans)

        wsgi_span = spans[0]

        # assert wsgi_span.t == '0000000000000001'
        # assert wsgi_span.p == '0000000000000001'
        assert wsgi_span.t == 1
        assert wsgi_span.p == 1

        assert 'X-INSTANA-T' in response.headers
        assert int(response.headers['X-INSTANA-T'], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert 'X-INSTANA-S' in response.headers
        assert int(response.headers['X-INSTANA-S'], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert 'X-INSTANA-L' in response.headers
        assert response.headers['X-INSTANA-L'] == '1'

        assert 'Server-Timing' in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers['Server-Timing'] == server_timing_value

    def test_response_header_capture(self) -> None:
        # Hack together a manual custom headers list
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", testenv["wsgi_server"] + "/response_headers"
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 200

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Synthetic
        assert not wsgi_span.sy
        assert not urllib3_span.sy
        assert not test_span.sy

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(testenv["wsgi_port"])
        assert wsgi_span.data["http"]["path"] == "/response_headers"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == "200"
        assert not wsgi_span.data["http"]["error"]

        # custom headers
        assert "X-Capture-This" in wsgi_span.data["http"]["header"]
        assert wsgi_span.data["http"]["header"]["X-Capture-This"] == "this"
        assert "X-Capture-That" in wsgi_span.data["http"]["header"]
        assert wsgi_span.data["http"]["header"]["X-Capture-That"] == "that"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_request_header_capture(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        request_headers = {
            "X-Capture-This-Too": "this too",
            "X-Capture-That-Too": "that too",
        }

        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", testenv["wsgi_server"] + "/", headers=request_headers
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response.status == 200

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Synthetic
        assert not wsgi_span.sy
        assert not urllib3_span.sy
        assert not test_span.sy

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not wsgi_span.ec

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(testenv["wsgi_port"])
        assert wsgi_span.data["http"]["path"] == "/"
        assert wsgi_span.data["http"]["method"] == "GET"
        assert wsgi_span.data["http"]["status"] == "200"
        assert not wsgi_span.data["http"]["error"]
        assert not wsgi_span.stack

        # custom headers
        assert "X-Capture-This-Too" in wsgi_span.data["http"]["header"]
        assert wsgi_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in wsgi_span.data["http"]["header"]
        assert wsgi_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

        agent.options.extra_http_headers = original_extra_http_headers



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

import re
from typing import Optional, Tuple

from opentelemetry.trace.span import (
    format_span_id,
    format_trace_id,
)

from instana.log import logger
from instana.util.ids import header_to_id, header_to_long_id

# See https://www.w3.org/TR/trace-context-2/#trace-flags for details on the bitmasks.
SAMPLED_BITMASK = 0b1;

class Traceparent:
    SPECIFICATION_VERSION = "00"
    TRACEPARENT_REGEX = re.compile(
        "^[0-9a-f][0-9a-e]-(?!0{32})([0-9a-f]{32})-(?!0{16})([0-9a-f]{16})-[0-9a-f]{2}"
    )

    def validate(self, traceparent):
        """
        Method used to validate the traceparent header
        :param traceparent: string
        :return: traceparent or None
        """
        try:
            if self.TRACEPARENT_REGEX.match(traceparent):
                return traceparent
        except Exception:
            logger.debug(
                "traceparent does not follow version {} specification".format(
                    self.SPECIFICATION_VERSION
                )
            )
        return None

    @staticmethod
    def get_traceparent_fields(traceparent: str) -> Tuple[Optional[str], Optional[int], Optional[int], Optional[bool]]:
        """
        Parses the validated traceparent header into its fields and returns the fields
        :param traceparent: the original validated traceparent header
        :return: version, trace_id, parent_id, sampled_flag
        """
        try:
            traceparent_properties = traceparent.split("-")
            version = traceparent_properties[0]
            trace_id = header_to_long_id(traceparent_properties[1])
            parent_id = header_to_id(traceparent_properties[2])
            flags = int(traceparent_properties[3], 16)
            sampled_flag = (flags & SAMPLED_BITMASK) == SAMPLED_BITMASK
            return version, trace_id, parent_id, sampled_flag
        except Exception as err:  # This method is intended to be called with a version 00 validated traceparent
            # This exception handling is added just for making sure we do not throw any unhandled exception
            # if somebody calls the method in the future without a validated traceparent
            logger.debug("Parsing the traceparent failed: {}".format(err))
            return None, None, None, None

    def update_traceparent(
        self,
        traceparent: Optional[str],
        in_trace_id: int,
        in_span_id: int,
        level: int,
    ) -> str:
        """
        This method updates the traceparent header or generates one if there was no traceparent incoming header or it
        was invalid
        :param traceparent: the original validated traceparent header
        :param in_trace_id: instana trace id, used when there is no preexisting trace_id from the traceparent header
        :param in_span_id: instana span id, used to update the parent id of the traceparent header
        :param level: instana level, used to determine the value of sampled flag of the traceparent header
        :return: the updated traceparent header
        """
        if (
            traceparent is None
        ):  # modify the trace_id part only when it was not present at all
            trace_id = (
                in_trace_id.zfill(32)
                if not isinstance(in_trace_id, int)
                else in_trace_id
            )
        else:
            # - We do not need the incoming upstream parent span ID for the header we sent downstream.
            # - We also do not care about the incoming version: The version field we sent downstream needs to match the
            #   format of the traceparent header we produce here, so we always send the version _we_ support downstream,
            #   even if the header coming from upstream supported a different version.
            # - Finally, we also do not care about the incoming sampled flag , we only need to communicate our own
            #   sampling decision downstream. The sampling decisions from our upstream is irrelevant for what we send
            #   downstream.
            _, trace_id, _, _ = self.get_traceparent_fields(traceparent)

        parent_id = (
            in_span_id.zfill(16) if not isinstance(in_span_id, int) else in_span_id
        )
        flags = level & SAMPLED_BITMASK
        flags = format(flags, "0>2x")

        if isinstance(trace_id, str):
            trace_id_out = trace_id
        else:
            trace_id_out = format_trace_id(trace_id)

        traceparent = f"{self.SPECIFICATION_VERSION}-{trace_id_out}-{format_span_id(parent_id)}-{flags}"
        return traceparent



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import uvicorn
from ...helpers import testenv
from instana.log import logger

testenv["fastapi_port"] = 10816
testenv["fastapi_server"] = "http://127.0.0.1:" + str(testenv["fastapi_port"])


def launch_fastapi():
    from .app import fastapi_server
    from instana.singletons import agent

    # Hack together a manual custom headers list; We'll use this in tests
    agent.options.extra_http_headers = [
        "X-Capture-This",
        "X-Capture-That",
        "X-Capture-This-Too",
        "X-Capture-That-Too",
    ]

    uvicorn.run(
        fastapi_server,
        host="127.0.0.1",
        port=testenv["fastapi_port"],
        log_level="critical",
    )



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import pytest
import urllib3
from typing import Generator

from instana.util.ids import hex_id
import tests.apps.pyramid.pyramid_app
from tests.helpers import testenv
from instana.singletons import tracer, agent
from instana.span.span import get_current_span


class TestPyramid:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """Clear all spans before a test run"""
        self.http = urllib3.PoolManager()
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

    def test_vanilla_requests(self) -> None:
        r = self.http.request("GET", testenv["pyramid_server"] + "/")
        assert r.status == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

    def test_get_request(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", testenv["pyramid_server"] + "/")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response.status == 200

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(pyramid_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(pyramid_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(pyramid_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert not get_current_span().is_recording()

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == pyramid_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert pyramid_span.p == urllib3_span.s

        # Synthetic
        assert not pyramid_span.sy
        assert not urllib3_span.sy
        assert not test_span.sy

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not pyramid_span.ec

        # wsgi
        assert pyramid_span.n == "wsgi"
        assert pyramid_span.data["http"]["host"] == "127.0.0.1:" + str(testenv["pyramid_port"])
        assert pyramid_span.data["http"]["url"] == "/"
        assert pyramid_span.data["http"]["method"] == "GET"
        assert pyramid_span.data["http"]["status"] == 200
        assert not pyramid_span.data["http"]["error"]
        assert pyramid_span.data["http"]["path_tpl"] == "/"

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert testenv["pyramid_server"] + "/" == urllib3_span.data["http"]["url"]
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

    def test_synthetic_request(self) -> None:
        headers = {"X-INSTANA-SYNTHETIC": "1"}

        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", testenv["pyramid_server"] + "/", headers=headers
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response.status == 200

        assert pyramid_span.sy
        assert not urllib3_span.sy
        assert not test_span.sy

    def test_500(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", testenv["pyramid_server"] + "/500")

        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response.status == 500

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(pyramid_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(pyramid_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(pyramid_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert not get_current_span().is_recording()

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert test_span.t == pyramid_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert pyramid_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert urllib3_span.ec == 1
        assert pyramid_span.ec == 1

        # wsgi
        assert pyramid_span.n == "wsgi"
        assert pyramid_span.data["http"]["host"] == "127.0.0.1:" + str(testenv["pyramid_port"])
        assert pyramid_span.data["http"]["url"] == "/500"
        assert pyramid_span.data["http"]["method"] == "GET"
        assert pyramid_span.data["http"]["status"] == 500
        assert pyramid_span.data["http"]["error"] == "internal error"
        assert pyramid_span.data["http"]["path_tpl"] == "/500"

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 500
        assert testenv["pyramid_server"] + "/500" == urllib3_span.data["http"]["url"]
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

    def test_exception(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", testenv["pyramid_server"] + "/exception"
            )

        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response.status == 500

        assert not get_current_span().is_recording()

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert test_span.t == pyramid_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert pyramid_span.p == urllib3_span.s

        # Error logging
        assert not test_span.ec
        assert urllib3_span.ec == 1
        assert pyramid_span.ec == 1

        # wsgi
        assert pyramid_span.n == "wsgi"
        assert pyramid_span.data["http"]["host"] == "127.0.0.1:" + str(testenv["pyramid_port"])
        assert pyramid_span.data["http"]["url"] == "/exception"
        assert pyramid_span.data["http"]["method"] == "GET"
        assert pyramid_span.data["http"]["status"] == 500
        assert pyramid_span.data["http"]["error"] == "fake exception"
        assert not pyramid_span.data["http"]["path_tpl"]

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 500
        assert (
            testenv["pyramid_server"] + "/exception" == urllib3_span.data["http"]["url"]
        )
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

    def test_response_header_capture(self) -> None:
        # Hack together a manual custom headers list
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", testenv["pyramid_server"] + "/response_headers"
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 200

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == pyramid_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert pyramid_span.p == urllib3_span.s

        # Synthetic
        assert not pyramid_span.sy
        assert not urllib3_span.sy
        assert not test_span.sy

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not pyramid_span.ec

        # wsgi
        assert pyramid_span.n == "wsgi"
        assert pyramid_span.data["http"]["host"] == "127.0.0.1:" + str(testenv["pyramid_port"])
        assert pyramid_span.data["http"]["url"] == "/response_headers"
        assert pyramid_span.data["http"]["method"] == "GET"
        assert pyramid_span.data["http"]["status"] == 200
        assert not pyramid_span.data["http"]["error"]
        assert pyramid_span.data["http"]["path_tpl"] == "/response_headers"

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert (
            testenv["pyramid_server"] + "/response_headers"
            == urllib3_span.data["http"]["url"]
        )
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # custom headers
        assert "X-Capture-This" in pyramid_span.data["http"]["header"]
        assert pyramid_span.data["http"]["header"]["X-Capture-This"] == "Ok"
        assert "X-Capture-That" in pyramid_span.data["http"]["header"]
        assert pyramid_span.data["http"]["header"]["X-Capture-That"] == "Ok too"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_request_header_capture(self) -> None:
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        request_headers = {
            "X-Capture-This-Too": "this too",
            "X-Capture-That-Too": "that too",
        }

        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", testenv["pyramid_server"] + "/", headers=request_headers
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response.status == 200

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == pyramid_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert pyramid_span.p == urllib3_span.s

        # Synthetic
        assert not pyramid_span.sy
        assert not urllib3_span.sy
        assert not test_span.sy

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not pyramid_span.ec

        # wsgi
        assert pyramid_span.n == "wsgi"
        assert pyramid_span.data["http"]["host"] == "127.0.0.1:" + str(testenv["pyramid_port"])
        assert pyramid_span.data["http"]["url"] == "/"
        assert pyramid_span.data["http"]["method"] == "GET"
        assert pyramid_span.data["http"]["status"] == 200
        assert not pyramid_span.data["http"]["error"]
        assert pyramid_span.data["http"]["path_tpl"] == "/"

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert testenv["pyramid_server"] + "/" == urllib3_span.data["http"]["url"]
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # custom headers
        assert "X-Capture-This-Too" in pyramid_span.data["http"]["header"]
        assert pyramid_span.data["http"]["header"]["X-Capture-This-Too"] == "this too"
        assert "X-Capture-That-Too" in pyramid_span.data["http"]["header"]
        assert pyramid_span.data["http"]["header"]["X-Capture-That-Too"] == "that too"

        agent.options.extra_http_headers = original_extra_http_headers

    def test_scrub_secret_path_template(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", testenv["pyramid_server"] + "/hello_user/oswald?secret=sshhh"
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        pyramid_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 200

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(pyramid_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(pyramid_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(pyramid_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert not get_current_span().is_recording()

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == pyramid_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert pyramid_span.p == urllib3_span.s

        # Synthetic
        assert not pyramid_span.sy
        assert not urllib3_span.sy
        assert not test_span.sy

        # Error logging
        assert not test_span.ec
        assert not urllib3_span.ec
        assert not pyramid_span.ec

        # wsgi
        assert pyramid_span.n == "wsgi"
        assert pyramid_span.data["http"]["host"] == "127.0.0.1:" + str(testenv["pyramid_port"])
        assert pyramid_span.data["http"]["url"] == "/hello_user/oswald"
        assert pyramid_span.data["http"]["method"] == "GET"
        assert pyramid_span.data["http"]["status"] == 200
        assert pyramid_span.data["http"]["params"] == "secret=<redacted>"
        assert not pyramid_span.data["http"]["error"]
        assert pyramid_span.data["http"]["path_tpl"] == "/hello_user/{user}"

        # urllib3
        assert test_span.data["sdk"]["name"] == "test"
        assert urllib3_span.n == "urllib3"
        assert urllib3_span.data["http"]["status"] == 200
        assert (
            testenv["pyramid_server"] + pyramid_span.data["http"]["url"]
            == urllib3_span.data["http"]["url"]
        )
        assert urllib3_span.data["http"]["method"] == "GET"
        assert urllib3_span.stack
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import math
import os
import time
import uuid
from typing import Any, Dict, Optional


class Profile(object):
    CATEGORY_CPU = "cpu"
    CATEGORY_MEMORY = "memory"
    CATEGORY_TIME = "time"
    TYPE_CPU_USAGE = "cpu-usage"
    TYPE_MEMORY_ALLOCATION_RATE = "memory-allocation-rate"
    TYPE_BLOCKING_CALLS = "blocking-calls"
    UNIT_NONE = ""
    UNIT_MILLISECOND = "millisecond"
    UNIT_MICROSECOND = "microsecond"
    UNIT_NANOSECOND = "nanosecond"
    UNIT_BYTE = "byte"
    UNIT_KILOBYTE = "kilobyte"
    UNIT_PERCENT = "percent"
    UNIT_SAMPLE = "sample"
    RUNTIME_PYTHON = "python"

    def __init__(
        self,
        category: str,
        typ: str,
        unit: str,
        roots: object,
        duration: int,
        timespan: int,
    ) -> None:
        self.process_id = str(os.getpid())
        self.id = generate_uuid()
        self.runtime = Profile.RUNTIME_PYTHON
        self.category = category
        self.type = typ
        self.unit = unit
        self.roots = roots
        self.duration = duration
        self.timespan = timespan
        self.timestamp = millis()

    def to_dict(self) -> Dict[str, Any]:
        profile_dict = {
            "pid": self.process_id,
            "id": self.id,
            "runtime": self.runtime,
            "category": self.category,
            "type": self.type,
            "unit": self.unit,
            "roots": [root.to_dict() for root in self.roots],
            "duration": self.duration,
            "timespan": self.timespan,
            "timestamp": self.timestamp,
        }

        return profile_dict


class CallSite:
    __slots__ = [
        "method_name",
        "file_name",
        "file_line",
        "measurement",
        "num_samples",
        "children",
    ]

    def __init__(self, method_name: str, file_name: str, file_line: int) -> None:
        self.method_name = method_name
        self.file_name = file_name
        self.file_line = file_line
        self.measurement: int = 0
        self.num_samples: int = 0
        self.children = dict()

    def create_key(self, method_name: str, file_name: str, file_line: int) -> str:
        return f"{method_name} ({file_name}:{file_line})"

    def find_child(
        self, method_name: str, file_name: str, file_line: int
    ) -> Optional[object]:
        key = self.create_key(method_name, file_name, file_line)
        if key in self.children:
            return self.children[key]

        return None

    def add_child(self, child: object) -> None:
        self.children[
            self.create_key(child.method_name, child.file_name, child.file_line)
        ] = child

    def remove_child(self, child: object) -> None:
        del self.children[
            self.create_key(child.method_name, child.file_name, child.file_line)
        ]

    def find_or_add_child(
        self, method_name: str, file_name: str, file_line: int
    ) -> object:
        child = self.find_child(method_name, file_name, file_line)
        if not child:
            child = CallSite(method_name, file_name, file_line)
            self.add_child(child)

        return child

    def increment(self, value: int, count: int) -> None:
        self.measurement += value
        self.num_samples += count

    def normalize(self, factor: int) -> None:
        self.measurement = self.measurement / factor
        self.num_samples = int(math.ceil(self.num_samples / factor))

        for child in self.children.values():
            child.normalize(factor)

    def floor(self) -> None:
        self.measurement = int(self.measurement)

        for child in self.children.values():
            child.floor()

    def to_dict(self) -> Dict[str, Any]:
        children_dicts = []
        for child in self.children.values():
            children_dicts.append(child.to_dict())

        call_site_dict = {
            "method_name": self.method_name,
            "file_name": self.file_name,
            "file_line": self.file_line,
            "measurement": self.measurement,
            "num_samples": self.num_samples,
            "children": children_dicts,
        }

        return call_site_dict


def millis() -> int:
    """
    Returns the current time in milliseconds since the Unix epoch (January 1, 1970).
    """
    return int(round(time.time() * 1000))


def generate_uuid() -> str:
    """
    Generates a UUID as string.
    """
    return str(uuid.uuid4())



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


import re
import wrapt
from typing import Any, Tuple, Dict, Callable

from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry import context, trace

from instana.log import logger
from instana.util.secrets import strip_secrets_from_query
from instana.singletons import agent, tracer
from instana.util.traceutils import extract_custom_headers
from instana.propagators.format import Format

import flask
from flask import request_started, request_finished, got_request_exception

path_tpl_re = re.compile("<.*>")


def request_started_with_instana(sender: flask.app.Flask, **extra: Any) -> None:
    try:
        env = flask.request.environ

        span_context = tracer.extract(Format.HTTP_HEADERS, env)

        span = tracer.start_span("wsgi", span_context=span_context)
        flask.g.span = span

        ctx = trace.set_span_in_context(span)
        token = context.attach(ctx)
        flask.g.token = token

        extract_custom_headers(span, env, format=True)

        span.set_attribute(SpanAttributes.HTTP_METHOD, flask.request.method)
        if "PATH_INFO" in env:
            span.set_attribute(SpanAttributes.HTTP_URL, env["PATH_INFO"])
        if "QUERY_STRING" in env and len(env["QUERY_STRING"]):
            scrubbed_params = strip_secrets_from_query(
                env["QUERY_STRING"],
                agent.options.secrets_matcher,
                agent.options.secrets_list,
            )
            span.set_attribute("http.params", scrubbed_params)
        if "HTTP_HOST" in env:
            span.set_attribute("http.host", env["HTTP_HOST"])

        if hasattr(flask.request.url_rule, "rule") and path_tpl_re.search(
            flask.request.url_rule.rule
        ):
            path_tpl = flask.request.url_rule.rule.replace("<", "{")
            path_tpl = path_tpl.replace(">", "}")
            span.set_attribute("http.path_tpl", path_tpl)
    except:
        logger.debug("Flask request_started_with_instana", exc_info=True)


def request_finished_with_instana(
    sender: flask.app.Flask, response: flask.wrappers.Response, **extra: Any
) -> None:
    span = None
    try:
        if not hasattr(flask.g, "span"):
            return

        span = flask.g.span
        if span:
            if 500 <= response.status_code:
                span.mark_as_errored()

            span.set_attribute(
                SpanAttributes.HTTP_STATUS_CODE, int(response.status_code)
            )
            extract_custom_headers(span, response.headers, format=False)

            tracer.inject(span.context, Format.HTTP_HEADERS, response.headers)
    except Exception:
        logger.debug("Flask request_finished_with_instana", exc_info=True)
    finally:
        if span and span.is_recording():
            span.end()


def log_exception_with_instana(
    sender: flask.app.Flask, exception: Exception, **extra: Any
) -> None:
    if hasattr(flask.g, "span") and flask.g.span:
        span = flask.g.span
        if span:
            span.record_exception(exception)
            # As of Flask 2.3.x:
            # https://github.com/pallets/flask/blob/
            # d0bf462866289ad8bfe29b6e4e1e0f531003ab34/src/flask/app.py#L1379
            # The `got_request_exception` signal, is only sent by
            # the `handle_exception` method which "always causes a 500"
            span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, 500)
            if span.is_recording():
                span.end()


def teardown_request_with_instana(*argv: Any, **kwargs: Any) -> None:
    """
    In the case of exceptions, request_finished_with_instana isn't called
    so we capture those cases here.
    """
    if hasattr(flask.g, "span") and flask.g.span:
        if len(argv) > 0 and argv[0]:
            span = flask.g.span
            span.record_exception(argv[0])
            if SpanAttributes.HTTP_STATUS_CODE not in span.attributes:
                span.set_attribute(SpanAttributes.HTTP_STATUS_CODE, 500)
        if flask.g.span.is_recording():
            flask.g.span.end()
        flask.g.span = None

    if hasattr(flask.g, "token") and flask.g.token:
        context.detach(flask.g.token)
        flask.g.token = None


@wrapt.patch_function_wrapper("flask", "Flask.full_dispatch_request")
def full_dispatch_request_with_instana(
    wrapped: Callable[..., flask.wrappers.Response],
    instance: flask.app.Flask,
    argv: Tuple,
    kwargs: Dict,
) -> flask.wrappers.Response:
    if not hasattr(instance, "_stan_wuz_here"):
        logger.debug(
            "Flask(blinker): Applying flask before/after instrumentation funcs"
        )
        setattr(instance, "_stan_wuz_here", True)
        got_request_exception.connect(log_exception_with_instana, instance)
        request_started.connect(request_started_with_instana, instance)
        request_finished.connect(request_finished_with_instana, instance)
        instance.teardown_request(teardown_request_with_instana)

    return wrapped(*argv, **kwargs)


logger.debug("Instrumenting flask (with blinker support)")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

from wsgiref.simple_server import make_server
from pyramid.config import Configurator
import logging

from pyramid.response import Response
import pyramid.httpexceptions as exc

from tests.helpers import testenv

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

testenv["pyramid_port"] = 10815
testenv["pyramid_server"] = "http://127.0.0.1:" + str(testenv["pyramid_port"])


def hello_world(request):
    return Response("Ok")


def please_fail(request):
    raise exc.HTTPInternalServerError("internal error")


def tableflip(request):
    raise BaseException("fake exception")


def response_headers(request):
    headers = {"X-Capture-This": "Ok", "X-Capture-That": "Ok too"}
    return Response("Stan wuz here with headers!", headers=headers)


def hello_user(request):
    user = request.matchdict["user"]
    return Response(f"Hello {user}!")


app = None
settings = {
    "pyramid.tweens": "tests.apps.pyramid.pyramid_utils.tweens.timing_tween_factory",
}
with Configurator(settings=settings) as config:
    config.add_route("hello", "/")
    config.add_view(hello_world, route_name="hello")
    config.add_route("fail", "/500")
    config.add_view(please_fail, route_name="fail")
    config.add_route("crash", "/exception")
    config.add_view(tableflip, route_name="crash")
    config.add_route("response_headers", "/response_headers")
    config.add_view(response_headers, route_name="response_headers")
    config.add_route("hello_user", "/hello_user/{user}")
    config.add_view(hello_user, route_name="hello_user")
    app = config.make_wsgi_app()

pyramid_server = make_server("127.0.0.1", testenv["pyramid_port"], app)



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import unittest
import urllib3
import flask
from unittest.mock import patch

from instana.util.ids import hex_id

if hasattr(flask.signals, 'signals_available'):
    from flask.signals import signals_available
else:
    # Beginning from 2.3.0 as stated in the notes
    # https://flask.palletsprojects.com/en/2.3.x/changes/#version-2-3-0
    # "Signals are always available. blinker>=1.6.2 is a required dependency.
    # The signals_available attribute is deprecated. #5056"
    signals_available = True

from opentelemetry.trace import SpanKind

import tests.apps.flask_app
from instana.singletons import tracer, agent
from instana.span.span import get_current_span
from tests.helpers import testenv


class TestFlask(unittest.TestCase):

    def setUp(self) -> None:
        """ Clear all spans before a test run """
        self.http = urllib3.PoolManager()
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()

    def tearDown(self) -> None:
        """ Do nothing for now """
        return None

    def test_vanilla_requests(self) -> None:
        r = self.http.request('GET', testenv["flask_server"] + '/')
        assert r.status == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

    def test_get_request(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request("GET", testenv["flask_server"] + "/")

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 200

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Synthetic
        assert wsgi_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert wsgi_span.ec is None

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert wsgi_span.data["http"]["host"] == "127.0.0.1:" + str(
            testenv["flask_port"]
        )
        assert "/" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 200 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 200 == urllib3_span.data["http"]["status"]
        assert testenv["flask_server"] + "/" == urllib3_span.data["http"]["url"]
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_get_request_with_query_params(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", testenv["flask_server"] + "/" + "?key1=val1&key2=val2"
            )

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 200

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Synthetic
        assert wsgi_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert wsgi_span.ec is None

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/" == wsgi_span.data["http"]["url"]
        assert wsgi_span.data["http"]["params"] == "key1=<redacted>&key2=<redacted>"
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 200 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 200 == urllib3_span.data["http"]["status"]
        assert testenv["flask_server"] + "/" == urllib3_span.data["http"]["url"]
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_get_request_with_suppression(self) -> None:
        headers = {'X-INSTANA-L':'0'}
        response = self.http.urlopen('GET', testenv["flask_server"] + '/', headers=headers)

        spans = self.recorder.queued_spans()

        assert response.headers.get("X-INSTANA-L", None) == "0"
        # The traceparent has to be present
        assert response.headers.get("traceparent", None) is not None
        # The last digit of the traceparent has to be 0
        assert response.headers["traceparent"][-1] == "0"

        # This should not be present
        assert response.headers.get("tracestate", None) is None

        # Assert that there are no spans in the recorded list
        assert spans == []

    def test_get_request_with_suppression_and_w3c(self) -> None:
        """Incoming Level 0 Plus W3C Trace Context Specification Headers"""
        headers = {
                'X-INSTANA-L':'0',
                'traceparent': '00-0af7651916cd43dd8448eb211c80319c-b9c7c989f97918e1-01',
                'tracestate': 'congo=ucfJifl5GOE,rojo=00f067aa0ba902b7'}

        response = self.http.urlopen('GET', testenv["flask_server"] + '/', headers=headers)

        spans = self.recorder.queued_spans()

        assert response.headers.get("X-INSTANA-L", None) == "0"
        # if X-INSTANA-L=0 then both X-INSTANA-T and X-INSTANA-S should not be present
        assert not response.headers.get("X-INSTANA-T", None)
        assert not response.headers.get("X-INSTANA-S", None)

        assert response.headers.get("traceparent", None) is not None
        assert response.headers["traceparent"].startswith("00-0af7651916cd43dd8448eb211c80319c")
        assert response.headers["traceparent"][-1] == "0"
        # The tracestate has to be present
        assert response.headers.get("tracestate", None) is not None

        # The 'in=' section can not be in the tracestate
        assert "in=" not in response.headers["tracestate"]

        # Assert that there are no spans in the recorded list
        assert spans == []

    def test_synthetic_request(self) -> None:
        headers = {
            'X-INSTANA-SYNTHETIC': '1'
        }

        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/', headers=headers)

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert wsgi_span.sy
        assert urllib3_span.sy is None
        assert test_span.sy is None

    def test_render_template(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/render')

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        render_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        assert response.status == 200

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == render_span.t
        assert test_span.t == urllib3_span.t
        assert test_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s
        assert render_span.p == wsgi_span.s

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert wsgi_span.ec is None
        assert render_span.ec is None

        # render
        assert "render" == render_span.n
        assert SpanKind.INTERNAL == render_span.k
        assert "flask_render_template.html" == render_span.data["render"]["name"]
        assert "template" == render_span.data["render"]["type"]
        assert render_span.data["log"]["message"] is None
        assert render_span.data["log"]["parameters"] is None

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/render" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 200 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 200 == urllib3_span.data["http"]["status"]
        assert testenv["flask_server"] + "/render" == urllib3_span.data["http"]["url"]
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_render_template_string(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/render_string')

        spans = self.recorder.queued_spans()
        assert len(spans) == 4

        render_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        assert response.status == 200

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == render_span.t
        assert test_span.t == urllib3_span.t
        assert test_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s
        assert render_span.p == wsgi_span.s

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert wsgi_span.ec is None
        assert render_span.ec is None

        # render
        assert "render" == render_span.n
        assert SpanKind.INTERNAL == render_span.k
        assert "(from string)" == render_span.data["render"]["name"]
        assert "template" == render_span.data["render"]["type"]
        assert render_span.data["log"]["message"] is None
        assert render_span.data["log"]["parameters"] is None

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/render_string" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 200 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 200 == urllib3_span.data["http"]["status"]
        assert (
            testenv["flask_server"] + "/render_string"
            == urllib3_span.data["http"]["url"]
        )
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_301(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/301', redirect=False)

        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert 301 == response.status

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert test_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert test_span.ec is None
        assert None == urllib3_span.ec
        assert None == wsgi_span.ec

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/301" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 301 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 301 == urllib3_span.data["http"]["status"]
        assert testenv["flask_server"] + "/301" == urllib3_span.data["http"]["url"]
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_custom_404(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/custom-404')

        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert 404 == response.status

        # assert 'X-INSTANA-T' in response.headers
        # assert int(response.headers['X-INSTANA-T']) == 16
        # assert response.headers['X-INSTANA-T'] == wsgi_span.t
        #
        # assert 'X-INSTANA-S' in response.headers
        # assert int(response.headers['X-INSTANA-S']) == 16
        # assert response.headers['X-INSTANA-S'] == wsgi_span.s
        #
        # assert 'X-INSTANA-L' in response.headers
        # assert response.headers['X-INSTANA-L'] == '1'
        #
        # assert 'Server-Timing' in response.headers
        # server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        # assert response.headers['Server-Timing'] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert test_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert test_span.ec is None
        assert None == urllib3_span.ec
        assert None == wsgi_span.ec

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/custom-404" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 404 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 404 == urllib3_span.data["http"]["status"]
        assert (
            testenv["flask_server"] + "/custom-404" == urllib3_span.data["http"]["url"]
        )
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_404(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/11111111111')

        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert 404 == response.status

        # assert 'X-INSTANA-T' in response.headers
        # assert int(response.headers['X-INSTANA-T']) == 16
        # assert response.headers['X-INSTANA-T'] == wsgi_span.t
        #
        # assert 'X-INSTANA-S' in response.headers
        # assert int(response.headers['X-INSTANA-S']) == 16
        # assert response.headers['X-INSTANA-S'] == wsgi_span.s
        #
        # assert 'X-INSTANA-L' in response.headers
        # assert response.headers['X-INSTANA-L'] == '1'
        #
        # assert 'Server-Timing' in response.headers
        # server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        # assert response.headers['Server-Timing'] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert test_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert test_span.ec is None
        assert None == urllib3_span.ec
        assert None == wsgi_span.ec

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/11111111111" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 404 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 404 == urllib3_span.data["http"]["status"]
        assert (
            testenv["flask_server"] + "/11111111111" == urllib3_span.data["http"]["url"]
        )
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_500(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/500')

        spans = self.recorder.queued_spans()

        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert 500 == response.status

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert test_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert test_span.ec is None
        assert 1 == urllib3_span.ec
        assert 1 == wsgi_span.ec

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/500" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 500 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 500 == urllib3_span.data["http"]["status"]
        assert testenv["flask_server"] + "/500" == urllib3_span.data["http"]["url"]
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_render_error(self) -> None:
        if signals_available is True:
            raise unittest.SkipTest("Exceptions without handlers vary with blinker")

        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/render_error')

        spans = self.recorder.queued_spans()

        assert len(spans) == 4

        log_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        assert 500 == response.status

        # assert 'X-INSTANA-T' in response.headers
        # assert int(response.headers['X-INSTANA-T']) == 16
        # assert response.headers['X-INSTANA-T'] == wsgi_span.t
        #
        # assert 'X-INSTANA-S' in response.headers
        # assert int(response.headers['X-INSTANA-S']) == 16
        # assert response.headers['X-INSTANA-S'] == wsgi_span.s
        #
        # assert 'X-INSTANA-L' in response.headers
        # assert response.headers['X-INSTANA-L'] == '1'
        #
        # assert 'Server-Timing' in response.headers
        # server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        # assert response.headers['Server-Timing'] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert test_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert test_span.ec is None
        assert 1 == urllib3_span.ec
        assert 1 == wsgi_span.ec

        # error log
        assert "log" == log_span.n
        assert log_span.data["log"]["message"] == "Exception on /render_error [GET]"
        assert (
            log_span.data["log"]["parameters"]
            == "<class 'jinja2.exceptions.TemplateSyntaxError'> unexpected '}'"
        )

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/render_error" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 500 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 500 == urllib3_span.data["http"]["status"]
        assert (
            testenv["flask_server"] + "/render_error" == urllib3_span.data["http"]["url"]
        )
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_exception(self) -> None:
        if signals_available is True:
            raise unittest.SkipTest("Exceptions without handlers vary with blinker")

        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/exception')

        spans = self.recorder.queued_spans()

        assert len(spans) == 4

        log_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        assert 500 == response.status

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert test_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s
        assert log_span.p == wsgi_span.s

        # Error logging
        assert test_span.ec is None
        assert 1 == urllib3_span.ec
        assert 1 == wsgi_span.ec
        assert 1 == log_span.ec

        # error log
        assert "log" == log_span.n
        assert log_span.data["log"]["message"] == "Exception on /exception [GET]"
        assert log_span.data["log"]["parameters"] == "<class 'Exception'> fake error"

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/exception" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 500 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 500 == urllib3_span.data["http"]["status"]
        assert testenv["flask_server"] + "/exception" == urllib3_span.data["http"]["url"]
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_custom_exception_with_log(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/exception-invalid-usage')

        spans = self.recorder.queued_spans()

        assert len(spans) == 4

        log_span = spans[0]
        wsgi_span = spans[1]
        urllib3_span = spans[2]
        test_span = spans[3]

        assert response
        assert 502 == response.status

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert test_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert test_span.ec is None
        assert 1 == urllib3_span.ec
        assert 1 == wsgi_span.ec
        assert 1 == log_span.ec

        # error log
        assert "log" == log_span.n
        assert log_span.data["log"]["message"] == "InvalidUsage error handler invoked"
        assert (
            log_span.data["log"]["parameters"]
            == "<class 'tests.apps.flask_app.app.InvalidUsage'> "
        )

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/exception-invalid-usage" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 502 == wsgi_span.data["http"]["status"]
        assert "Simulated custom exception" == wsgi_span.data["http"]["error"]
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 502 == urllib3_span.data["http"]["status"]
        assert (
            testenv["flask_server"] + "/exception-invalid-usage"
            == urllib3_span.data["http"]["url"]
        )
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should NOT have a path template for this route
        assert wsgi_span.data["http"]["path_tpl"] is None

    def test_path_templates(self) -> None:
        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/users/Ricky/sayhello')

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 200

        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert wsgi_span.ec is None

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/users/Ricky/sayhello" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 200 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 200 == urllib3_span.data["http"]["status"]
        assert (
            testenv["flask_server"] + "/users/Ricky/sayhello"
            == urllib3_span.data["http"]["url"]
        )
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # We should have a reported path template for this route
        assert "/users/{username}/sayhello" == wsgi_span.data["http"]["path_tpl"]

    def test_request_header_capture(self) -> None:
        # Hack together a manual custom headers list
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This-Too", "X-Capture-That-Too"]

        request_headers = {
            "X-Capture-This-Too": "this too",
            "X-Capture-That-Too": "that too",
        }

        with tracer.start_as_current_span("test"):
            response = self.http.request(
                "GET", testenv["flask_server"] + "/", headers=request_headers
            )

        assert response
        assert response.status == 200

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        assert wsgi_span.ec is None
        assert wsgi_span.stack is None

        assert "/" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 200 == wsgi_span.data["http"]["status"]

        assert "X-Capture-This-Too" in wsgi_span.data["http"]["header"]
        assert "this too" == wsgi_span.data["http"]["header"]["X-Capture-This-Too"]
        assert "X-Capture-That-Too" in wsgi_span.data["http"]["header"]
        assert "that too" == wsgi_span.data["http"]["header"]["X-Capture-That-Too"]

        agent.options.extra_http_headers = original_extra_http_headers


    def test_response_header_capture(self) -> None:
        # Hack together a manual custom headers list
        original_extra_http_headers = agent.options.extra_http_headers
        agent.options.extra_http_headers = ["X-Capture-This", "X-Capture-That"]

        with tracer.start_as_current_span("test"):
            response = self.http.request('GET', testenv["flask_server"] + '/response_headers')

        spans = self.recorder.queued_spans()
        assert len(spans) == 3

        wsgi_span = spans[0]
        urllib3_span = spans[1]
        test_span = spans[2]

        assert response
        assert response.status == 200
        assert "X-INSTANA-T" in response.headers
        assert int(response.headers["X-INSTANA-T"], 16)
        assert response.headers["X-INSTANA-T"] == hex_id(wsgi_span.t)

        assert "X-INSTANA-S" in response.headers
        assert int(response.headers["X-INSTANA-S"], 16)
        assert response.headers["X-INSTANA-S"] == hex_id(wsgi_span.s)

        assert "X-INSTANA-L" in response.headers
        assert response.headers["X-INSTANA-L"] == "1"

        assert "Server-Timing" in response.headers
        server_timing_value = f"intid;desc={hex_id(wsgi_span.t)}"
        assert response.headers["Server-Timing"] == server_timing_value

        assert get_current_span().is_recording() is False

        # Same traceId
        assert test_span.t == urllib3_span.t
        assert urllib3_span.t == wsgi_span.t

        # Parent relationships
        assert urllib3_span.p == test_span.s
        assert wsgi_span.p == urllib3_span.s

        # Synthetic
        assert wsgi_span.sy is None
        assert urllib3_span.sy is None
        assert test_span.sy is None

        # Error logging
        assert test_span.ec is None
        assert urllib3_span.ec is None
        assert wsgi_span.ec is None

        # urllib3
        assert "test" == test_span.data["sdk"]["name"]
        assert "urllib3" == urllib3_span.n
        assert 200 == urllib3_span.data["http"]["status"]
        assert (
            testenv["flask_server"] + "/response_headers"
            == urllib3_span.data["http"]["url"]
        )
        assert "GET" == urllib3_span.data["http"]["method"]
        assert urllib3_span.stack is not None
        assert type(urllib3_span.stack) is list
        assert len(urllib3_span.stack) > 1

        # wsgi
        assert "wsgi" == wsgi_span.n
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/response_headers" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert 200 == wsgi_span.data["http"]["status"]
        assert wsgi_span.data["http"]["error"] is None
        assert wsgi_span.stack is None

        assert "X-Capture-This" in wsgi_span.data["http"]["header"]
        assert "Ok" == wsgi_span.data["http"]["header"]["X-Capture-This"]
        assert "X-Capture-That" in wsgi_span.data["http"]["header"]

        assert "Ok too" == wsgi_span.data["http"]["header"]["X-Capture-That"]

        agent.options.extra_http_headers = original_extra_http_headers

    def test_request_started_exception(self) -> None:
        with tracer.start_as_current_span("test"):
            with patch(
                "instana.singletons.tracer.extract",
                side_effect=Exception("mocked error"),
            ):
                self.http.request("GET", testenv["flask_server"] + "/")

        spans = self.recorder.queued_spans()
        assert len(spans) == 2

    @unittest.skipIf(
        not signals_available,
        "log_exception_with_instana needs to be covered only with blinker",
    )
    def test_got_request_exception(self) -> None:
        response = self.http.request(
            "GET", testenv["flask_server"] + "/got_request_exception"
        )

        spans = self.recorder.queued_spans()
        assert len(spans) == 1

        wsgi_span = spans[0]

        assert response
        assert 500 == response.status

        assert get_current_span().is_recording() is False

        # Error logging
        assert wsgi_span.ec == 1

        # wsgi
        assert wsgi_span.n == "wsgi"
        assert (
            "127.0.0.1:" + str(testenv["flask_port"]) == wsgi_span.data["http"]["host"]
        )
        assert "/got_request_exception" == wsgi_span.data["http"]["url"]
        assert "GET" == wsgi_span.data["http"]["method"]
        assert wsgi_span.data["http"]["status"] == 500
        assert wsgi_span.data["http"]["error"] == "RuntimeError()"
        assert wsgi_span.stack is None



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import threading
from typing import Generator

import pytest

from instana.autoprofile.profiler import Profiler
from instana.autoprofile.runtime import RuntimeInfo


class TestProfiler:
    @pytest.fixture(autouse=True)
    def _resources(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Create a new Profiler.
        self.profiler = Profiler(None)
        self.profiler.start(disable_timers=True)
        yield
        # teardown
        self.profiler.destroy()

    def test_run_in_main_thread(self) -> None:
        if RuntimeInfo.OS_WIN:
            return

        result = {}

        def _run():
            result["thread_id"] = threading.current_thread().ident

        def _thread():
            self.profiler.run_in_main_thread(_run)

        t = threading.Thread(target=_thread)
        t.start()
        t.join()

        assert threading.current_thread().ident == result["thread_id"]



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020

import logging
from typing import Generator
from unittest.mock import patch

import pytest
from opentelemetry.trace import SpanKind

from instana.singletons import agent, tracer


class TestLogging:
    @pytest.fixture(autouse=True)
    def _resource(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        # Clear all spans before a test run
        self.recorder = tracer.span_processor
        self.recorder.clear_spans()
        self.logger = logging.getLogger("unit test")
        yield
        # tearDown
        # Ensure that allow_exit_as_root has the default value
        agent.options.allow_exit_as_root = False

    def test_no_span(self) -> None:
        self.logger.setLevel(logging.INFO)
        with tracer.start_as_current_span("test"):
            self.logger.info("info message")

        spans = self.recorder.queued_spans()

        assert len(spans) == 1

    def test_extra_span(self) -> None:
        with tracer.start_as_current_span("test"):
            self.logger.warning("foo %s", "bar")

        spans = self.recorder.queued_spans()

        assert len(spans) == 2
        assert spans[0].k is SpanKind.CLIENT
        assert spans[0].data["log"].get("message") == "foo bar"

    def test_log_with_tuple(self) -> None:
        with tracer.start_as_current_span("test"):
            self.logger.warning("foo %s", ("bar",))

        spans = self.recorder.queued_spans()

        assert len(spans) == 2
        assert spans[0].k is SpanKind.CLIENT
        assert spans[0].data["log"].get("message") == "foo ('bar',)"

    def test_log_with_dict(self) -> None:
        with tracer.start_as_current_span("test"):
            self.logger.warning("foo %s", {"bar": 18})

        spans = self.recorder.queued_spans()

        assert len(spans) == 2
        assert spans[0].k is SpanKind.CLIENT
        assert spans[0].data["log"].get("message") == "foo {'bar': 18}"

    def test_parameters(self) -> None:
        with tracer.start_as_current_span("test"):
            try:
                a = 42
                b = 0
                c = a / b
            except Exception as e:
                self.logger.exception("Exception: %s", str(e))

        spans = self.recorder.queued_spans()

        assert len(spans) == 2
        assert spans[0].data["log"].get("parameters") is not None

    def test_no_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        self.logger.info("info message")

        spans = self.recorder.queued_spans()

        assert len(spans) == 0

    def test_root_exit_span(self) -> None:
        agent.options.allow_exit_as_root = True
        self.logger.warning("foo %s", "bar")

        spans = self.recorder.queued_spans()

        assert len(spans) == 1
        assert spans[0].k is SpanKind.CLIENT
        assert spans[0].data["log"].get("message") == "foo bar"

    def test_exception(self) -> None:
        with tracer.start_as_current_span("test"):
            with patch(
                "instana.span.span.InstanaSpan.add_event",
                side_effect=Exception("mocked error"),
            ):
                self.logger.warning("foo %s", "bar")

        spans = self.recorder.queued_spans()

        assert len(spans) == 2
        assert spans[0].k is SpanKind.CLIENT
        assert spans[0].data["log"] == {}

    def test_log_caller(self, caplog: pytest.LogCaptureFixture) -> None:
        handler = logging.StreamHandler()
        handler.setFormatter(
            logging.Formatter("source: %(funcName)s, message: %(message)s")
        )
        self.logger.addHandler(handler)

        def log_custom_warning():
            self.logger.warning("foo %s", "bar")

        with tracer.start_as_current_span("test"):
            log_custom_warning()

        assert caplog.records[-1].funcName == "log_custom_warning"

        self.logger.removeHandler(handler)

    @pytest.mark.parametrize(
        "stacklevel, expected_caller_name",
        [
            (1, "log_custom_warning"),
            (2, "main"),
        ],
    )
    def test_log_caller_with_stacklevel(
        self,
        caplog: pytest.LogCaptureFixture,
        stacklevel: int,
        expected_caller_name: str,
    ) -> None:
        handler = logging.StreamHandler()
        handler.setFormatter(
            logging.Formatter("source: %(funcName)s, message: %(message)s")
        )
        self.logger.addHandler(handler)

        def log_custom_warning():
            self.logger.warning("foo %s", "bar", stacklevel=stacklevel)

        def main():
            log_custom_warning()

        with tracer.start_as_current_span("test"):
            main()

        assert caplog.records[-1].funcName == expected_caller_name

        self.logger.removeHandler(handler)

        spans = self.recorder.queued_spans()
        assert len(spans) == 2
        assert spans[0].k is SpanKind.CLIENT

        assert spans[0].data["log"].get("message") == "foo bar"



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2020


from instana.span.span import InstanaSpan
from instana.log import logger
from instana.util.traceutils import get_tracer_tuple, tracing_is_off

try:
    import pymongo
    from bson import json_util
    from opentelemetry.semconv.trace import SpanAttributes

    class MongoCommandTracer(pymongo.monitoring.CommandListener):
        def __init__(self) -> None:
            self.__active_commands = {}

        def started(self, event: pymongo.monitoring.CommandStartedEvent) -> None:
            tracer, parent_span, _ = get_tracer_tuple()
            # return early if we're not tracing
            if tracing_is_off():
                return
            parent_context = parent_span.get_span_context() if parent_span else None

            with tracer.start_as_current_span(
                "mongo", span_context=parent_context
            ) as span:
                self._collect_connection_tags(span, event)
                self._collect_command_tags(span, event)

                # include collection name into the namespace if provided
                if event.command_name in event.command:
                    span.set_attribute(
                        SpanAttributes.DB_MONGODB_COLLECTION,
                        event.command.get(event.command_name),
                    )

                self.__active_commands[event.request_id] = span

        def succeeded(self, event: pymongo.monitoring.CommandStartedEvent) -> None:
            active_span = self.__active_commands.pop(event.request_id, None)

            # return early if we're not tracing
            if active_span is None:
                return

        def failed(self, event: pymongo.monitoring.CommandStartedEvent) -> None:
            active_span = self.__active_commands.pop(event.request_id, None)

            # return early if we're not tracing
            if active_span is None:
                return

            active_span.log_exception(event.failure)

        def _collect_connection_tags(
            self, span: InstanaSpan, event: pymongo.monitoring.CommandStartedEvent
        ) -> None:
            (host, port) = event.connection_id

            span.set_attribute(SpanAttributes.SERVER_ADDRESS, host)
            span.set_attribute(SpanAttributes.SERVER_PORT, str(port))
            span.set_attribute(SpanAttributes.DB_NAME, event.database_name)

        def _collect_command_tags(self, span, event) -> None:
            """
            Extract MongoDB command name and arguments and attach it to the span
            """
            cmd = event.command_name
            span.set_attribute("command", cmd)

            for key in ["filter", "query"]:
                if key in event.command:
                    span.set_attribute(
                        "filter", json_util.dumps(event.command.get(key))
                    )
                    break

            # The location of command documents within the command object depends on the name
            # of this command. This is the name -> command object key mapping
            cmd_doc_locations = {
                "insert": "documents",
                "update": "updates",
                "delete": "deletes",
                "aggregate": "pipeline",
            }

            cmd_doc = None
            if cmd in cmd_doc_locations:
                cmd_doc = event.command.get(cmd_doc_locations[cmd])
            elif (
                cmd.lower() == "mapreduce"
            ):  # mapreduce command was renamed to mapReduce in pymongo 3.9.0
                # mapreduce command consists of two mandatory parts: map and reduce
                cmd_doc = {
                    "map": event.command.get("map"),
                    "reduce": event.command.get("reduce"),
                }

            if cmd_doc is not None:
                span.set_attribute("json", json_util.dumps(cmd_doc))

    pymongo.monitoring.register(MongoCommandTracer())

    logger.debug("Instrumenting pymongo")

except ImportError:
    pass



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2019


import sys
import wrapt
import logging
from collections.abc import Mapping
from typing import Any, Tuple, Dict, Callable

from instana.log import logger
from instana.util.traceutils import get_tracer_tuple, tracing_is_off


@wrapt.patch_function_wrapper("logging", "Logger._log")
def log_with_instana(
    wrapped: Callable[..., None],
    instance: logging.Logger,
    argv: Tuple[int, str, Tuple[Any, ...]],
    kwargs: Dict[str, Any],
) -> Callable[..., None]:
    # argv[0] = level
    # argv[1] = message
    # argv[2] = args for message

    # We take into consideration if `stacklevel` is already present in `kwargs`.
    # This prevents the error `_log() got multiple values for keyword argument 'stacklevel'`
    stacklevel_in = kwargs.pop("stacklevel", 1)
    stacklevel = stacklevel_in + 1 + (sys.version_info >= (3, 14))

    try:
        # Only needed if we're tracing and serious log
        if tracing_is_off() or argv[0] < logging.WARN:
            return wrapped(*argv, **kwargs, stacklevel=stacklevel)

        tracer, parent_span, _ = get_tracer_tuple()

        msg = str(argv[1])
        args = argv[2]
        if args and len(args) == 1 and isinstance(args[0], Mapping) and args[0]:
            args = args[0]

        # get the formatted log message
        msg = msg % args

        # get additional information if an exception is being handled
        parameters = None
        (t, v, tb) = sys.exc_info()
        if t is not None and v is not None:
            parameters = "{} {}".format(t, v)

        parent_context = parent_span.get_span_context() if parent_span else None

        # create logging span
        with tracer.start_as_current_span("log", span_context=parent_context) as span:
            event_attributes = {"message": msg}
            if parameters is not None:
                event_attributes.update({"parameters": parameters})
            span.add_event(name="log_with_instana", attributes=event_attributes)
            # extra tags for an error
            if argv[0] >= logging.ERROR:
                span.mark_as_errored()

    except Exception:
        logger.debug("log_with_instana:", exc_info=True)

    return wrapped(*argv, **kwargs, stacklevel=stacklevel)


logger.debug("Instrumenting logging")



# (c) Copyright IBM Corp. 2021
# (c) Copyright Instana Inc. 2021

from typing import Generator

import pytest
from opentelemetry.trace import (
    format_span_id,
    format_trace_id,
)

from instana.propagators.binary_propagator import BinaryPropagator
from instana.span_context import SpanContext
from instana.util.ids import hex_id


class TestBinaryPropagator:
    @pytest.fixture(autouse=True)
    def _resources(self) -> Generator[None, None, None]:
        """SetUp and TearDown"""
        # setup
        self.bp = BinaryPropagator()
        yield

    def test_inject_carrier_dict(self, trace_id: int, span_id: int, hex_trace_id: str, hex_span_id: str) -> None:
        carrier = {}
        ctx = SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            is_remote=False,
            level=1,
            baggage={},
            sampled=True,
            synthetic=False,
        )
        carrier = self.bp.inject(ctx, carrier)

        assert carrier[b"x-instana-t"] == hex_trace_id.encode("utf-8")
        assert carrier[b"x-instana-s"] == hex_span_id.encode("utf-8")
        assert carrier[b"x-instana-l"] == b"1"
        assert carrier[b"server-timing"] == f"intid;desc={hex_id(trace_id)}".encode("utf-8")

    def test_inject_carrier_dict_w3c_True(self, trace_id: int, span_id: int, hex_trace_id: str, hex_span_id: str) -> None:
        carrier = {}
        ctx = SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            is_remote=False,
            level=1,
            baggage={},
            sampled=True,
            synthetic=False,
        )
        carrier = self.bp.inject(ctx, carrier, disable_w3c_trace_context=False)

        assert carrier[b"x-instana-t"] == hex_trace_id.encode("utf-8")
        assert carrier[b"x-instana-s"] == hex_span_id.encode("utf-8")
        assert carrier[b"x-instana-l"] == b"1"
        assert carrier[b"server-timing"] == f"intid;desc={hex_id(trace_id)}".encode("utf-8")
        assert carrier[
            b"traceparent"
        ] == f"00-{format_trace_id(trace_id)}-{format_span_id(span_id)}-01".encode(
            "utf-8"
        )
        assert carrier[b"tracestate"] == f"in={hex_id(trace_id)};{hex_id(span_id)}".encode("utf-8")

    def test_inject_carrier_list(self, trace_id: int, span_id: int, hex_trace_id: str, hex_span_id: str) -> None:
        carrier = []
        ctx = SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            is_remote=False,
            level=1,
            baggage={},
            sampled=True,
            synthetic=False,
        )
        carrier = self.bp.inject(ctx, carrier)

        assert isinstance(carrier, list)
        assert carrier[0] == (b"x-instana-t", hex_trace_id.encode("utf-8"))
        assert carrier[1] == (b"x-instana-s", hex_span_id.encode("utf-8"))
        assert carrier[2] == (b"x-instana-l", b"1")
        assert carrier[3] == (
            b"server-timing",
            f"intid;desc={hex_id(trace_id)}".encode("utf-8"),
        )

    def test_inject_carrier_list_w3c_True(self, trace_id: int, span_id: int, hex_trace_id: str, hex_span_id: str) -> None:
        carrier = []
        ctx = SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            is_remote=False,
            level=1,
            baggage={},
            sampled=True,
            synthetic=False,
        )
        carrier = self.bp.inject(ctx, carrier, disable_w3c_trace_context=False)

        assert isinstance(carrier, list)
        assert carrier[0] == (
            b"traceparent",
            f"00-{format_trace_id(trace_id)}-{format_span_id(span_id)}-01".encode(
                "utf-8"
            ),
        )
        assert carrier[1] == (
            b"tracestate",
            f"in={hex_id(trace_id)};{hex_id(span_id)}".encode("utf-8"),
        )
        assert carrier[2] == (b"x-instana-t", hex_trace_id.encode("utf-8"))
        assert carrier[3] == (b"x-instana-s", hex_span_id.encode("utf-8"))
        assert carrier[4] == (b"x-instana-l", b"1")
        assert carrier[5] == (
            b"server-timing",
            f"intid;desc={hex_id(trace_id)}".encode("utf-8"),
        )

    def test_inject_carrier_tuple(self, trace_id: int, span_id: int, hex_trace_id: str, hex_span_id: str) -> None:
        carrier = ()
        ctx = SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            is_remote=False,
            level=1,
            baggage={},
            sampled=True,
            synthetic=False,
        )
        carrier = self.bp.inject(ctx, carrier)

        assert isinstance(carrier, tuple)
        assert carrier[0] == (b"x-instana-t", hex_trace_id.encode("utf-8"))
        assert carrier[1] == (b"x-instana-s", hex_span_id.encode("utf-8"))
        assert carrier[2] == (b"x-instana-l", b"1")
        assert carrier[3] == (
            b"server-timing",
            f"intid;desc={hex_id(trace_id)}".encode("utf-8"),
        )

    def test_inject_carrier_tuple_w3c_True(self, trace_id: int, span_id: int, hex_trace_id: str, hex_span_id: str) -> None:
        carrier = ()
        ctx = SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            is_remote=False,
            level=1,
            baggage={},
            sampled=True,
            synthetic=False,
        )
        carrier = self.bp.inject(ctx, carrier, disable_w3c_trace_context=False)

        assert isinstance(carrier, tuple)
        assert carrier[0] == (
            b"traceparent",
            f"00-{format_trace_id(trace_id)}-{format_span_id(span_id)}-01".encode(
                "utf-8"
            ),
        )
        assert carrier[1] == (
            b"tracestate",
            f"in={hex_id(trace_id)};{hex_id(span_id)}".encode("utf-8"),
        )
        assert carrier[2] == (b"x-instana-t", hex_trace_id.encode("utf-8"))
        assert carrier[3] == (b"x-instana-s", hex_span_id.encode("utf-8"))
        assert carrier[4] == (b"x-instana-l", b"1")
        assert carrier[5] == (
            b"server-timing",
            f"intid;desc={hex_id(trace_id)}".encode("utf-8"),
        )

    def test_inject_carrier_set_exception(self, trace_id: int, span_id: int) -> None:
        carrier = set()
        ctx = SpanContext(
            span_id=span_id,
            trace_id=trace_id,
            is_remote=False,
            level=1,
            baggage={},
            sampled=True,
            synthetic=False,
        )
        carrier = self.bp.inject(ctx, carrier)
        assert not carrier
